/****************************************************************************
   QGnokii - an alternative frontend for Gnokii
    Copyright (C) 2006  Piotr Jedryczek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
**************************************************************************/


#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <gnokii.h>
#include <strings.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <limits.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>
#include <syslog.h>
#include <sys/utsname.h>

/*******************************************************************************
* Meaning of codes:
* 00 - begin code
* 97 - "this function is currently broken on this system" code
* 98 - error code
* 99 - end code
* 01 - identify request
* 02 - send SMS request
* 03 - get contents of the phonebook request
* 04 - get ToDo list request
* 05 - purge ToDo list request
* 06 - delete phonebook entry
* 07 - write phonebook entry
* all other - request code
*******************************************************************************/

#define BEGIN_CODE "00"
#define END_CODE "99"
#define ERR_CODE "98"
#define IDENTIFY_CODE "01"
#define SENDSMS_CODE "02"
#define GETPB_CODE "03"
#define GETTODO_CODE "04"
#define PURGETODO_CODE "05"
#define DELPB_CODE "06"
#define WRITEPB_CODE "07"
#define CODE_LEN 3
#define MAXLINE 1024
#define OK_MSG "OK"
#define SYS_OK "OK"
#define SYS_WRONG "WR"

#ifndef false
#  define false (0)
#endif

#ifndef true
#  define true (!false)
#endif

#ifndef bool
#  define bool int
#endif


static struct gn_statemachine *state;
static gn_data *data;
static gn_error *error;
char reqno[3];
int listenfd, connfd;
struct sockaddr_in servaddr, cliaddr;
struct utsname sysname;

int daemon_init(void)
{
	pid_t pid;
	
	if((pid = fork()) < 0)
		return(-1);
	else if(pid != 0)
		exit(0);
	setsid();
	chdir("/");
	umask(0);
	return(0);
}
static void busterminate(void)
{
        gn_lib_phone_close(state);
        gn_lib_phoneprofile_free(&state);
        gn_lib_library_free();
}

void got_sigterm(int foo)
{
        busterminate();
	closelog();
	close(connfd);
        return;
}

static void businit(void)
{
	if (gn_lib_phoneprofile_load_from_file(NULL, NULL, &state) != GN_ERR_NONE)
		exit(2);

        /* register cleanup function */
        atexit(busterminate);

        if (GN_ERR_NONE != gn_lib_phone_open(state))
                exit(2);
        data = &state->sm_data;
}

void phone_identify()
{
	char imei[64], model[64], revision[64], manufacturer[64];
	char data[261];
	int i;

	/* Initialize phone bus */
	businit();

	/* Send the data to client */
	write(connfd, BEGIN_CODE, CODE_LEN);
	usleep(100);
	write(connfd, reqno, CODE_LEN);
	strcpy(imei, gn_lib_get_phone_imei(state));
	strcpy(manufacturer, gn_lib_get_phone_manufacturer(state));
	strcpy(model, gn_lib_get_phone_model(state));
	strcpy(revision, gn_lib_get_phone_revision(state));
	sprintf(data, "%s;%s;%s;%s;", imei, revision, model, manufacturer);
	write(connfd, data, strlen(data)+1);
	sleep(1);
	write(connfd, END_CODE, CODE_LEN);
	
	/* Terminate phone bus */
	busterminate();
}
void send_sms()
{
	gn_sms sms;
	gn_error error;
	char number[16], smsc[16], delreport[2], ifsmsc[2];
	char msgbody[161];

	businit();

	/* Set the default values */
	strcpy(delreport, "0");
	strcpy(ifsmsc, "1");
	
	/* Send confirmation to client */
	write(connfd, BEGIN_CODE, CODE_LEN);
	write(connfd, reqno, CODE_LEN);

	/* Get data from client */
	read(connfd, number, sizeof(number));
	read(connfd, ifsmsc, sizeof(ifsmsc));
	if(strcmp(ifsmsc, "1") == 0)
		read(connfd, smsc, sizeof(smsc));
	read(connfd, delreport, sizeof(delreport));
	read(connfd, msgbody, sizeof(msgbody));

	/* The memory is zeroed here */
        gn_sms_default_submit(&sms);

        memset(&sms.remote.number, 0, sizeof(sms.remote.number));
        strncpy(sms.remote.number, number, sizeof(sms.remote.number) - 1);
        if (sms.remote.number[0] == '+')
                sms.remote.type = GN_GSM_NUMBER_International;
        else
                sms.remote.type = GN_GSM_NUMBER_Unknown;
	
	/* If ifsmsc == 1, set SMS Center number */
	if(strcmp(ifsmsc, "1") == 0)
	{
		strncpy(sms.smsc.number, smsc, sizeof(sms.smsc.number) - 1);
        		if (sms.smsc.number[0] == '+')
				sms.smsc.type = GN_GSM_NUMBER_International;
			else
				sms.smsc.type = GN_GSM_NUMBER_Unknown;
	}
	else if(strcmp(ifsmsc, "0") == 0)
	{	
		/* If not, get it from phone memory */
		data->message_center = calloc(1, sizeof(gn_sms_message_center));
		data->message_center->id = 1;
		if(data->message_center->id < 1 || data->message_center->id > 5)
		{
			free(data->message_center);
			write(connfd, ERR_CODE, CODE_LEN);
			syslog(LOG_ERR, "Msg center error");
			return;
		}
		if(gn_sm_functions(GN_OP_GetSMSCenter, data, state) == GN_ERR_NONE) 
		{
			strcpy(sms.smsc.number, data->message_center->smsc.number);
			sms.smsc.type = data->message_center->smsc.type;
		}
		free(data->message_center);
	}
	else
		syslog(LOG_ERR, "Inappropriate request from client");

	if(!sms.smsc.type) sms.smsc.type = GN_GSM_NUMBER_Unknown;

	/* Delivery report */
	if(strcmp(delreport, "1") == 0)
		sms.delivery_report = true;
	
	/* Message body */
        strncpy(sms.user_data[0].u.text, msgbody, sizeof(sms.user_data));
        sms.user_data[0].length = strlen(msgbody)+1;
	sms.user_data[0].type = GN_SMS_DATA_Text;
	
	data->sms = &sms;

	/* Send the message */
	error = gn_sms_send(data, state);
	if(error == GN_ERR_NONE)
		write(connfd, END_CODE, CODE_LEN);
	else
	{
		write(connfd, ERR_CODE, CODE_LEN);
		syslog(LOG_ERR, "Sending error, libgnokii error %d", error);
	}
	
	busterminate();
}                

void get_pb()
{
	gn_phonebook_entry entry;
	gn_memory_status memstat;
	gn_error error;
	char memtype[3];
	char howmany_str[8];
	int howmany;
	int i;
	
	/* Initialize phone bus */
	businit();

	/* Send confirmation to client */
	write(connfd, BEGIN_CODE, CODE_LEN);
	write(connfd, reqno, CODE_LEN);
	
	/* Check OS name */
	if(strcmp(sysname.sysname, "OpenBSD") == 0)
	{
		write(connfd, SYS_WRONG, CODE_LEN);
		busterminate();
		return;
	}
	else
		write(connfd, SYS_OK, CODE_LEN);
	
	/* Get the data from client */
	read(connfd, memtype, sizeof(memtype));
	entry.memory_type = gn_str2memory_type(memtype);

	/* Get memory status */
	data->memory_status = &memstat;
        memstat.memory_type = entry.memory_type;
	if((error = gn_sm_functions(GN_OP_GetMemoryStatus, data, state)) == GN_ERR_NONE) 
		howmany = memstat.used;
	else	
	{
		write(connfd, ERR_CODE, CODE_LEN);
		syslog(LOG_ERR, "Can't get memory status");
		return;
	}
	
	/* Send information about length to client */
	snprintf(howmany_str, 8, "%d", howmany);
	write(connfd, howmany_str, strlen(howmany_str)+1);

	/* Get entries and send them to client */
	for(i = 1; i <= howmany; i++)
	{
		entry.location = i;
		data->phonebook_entry = &entry;
		error = gn_sm_functions(GN_OP_ReadPhonebook, data, state);
		if(error == GN_ERR_NONE)
		{
			usleep(6000);
			write(connfd, entry.name, sizeof(entry.name));
			usleep(6000);
			write(connfd, entry.number, sizeof(entry.number));
		}
		else if(error == GN_ERR_EMPTYLOCATION)
		{
			write(connfd, "Empty name", strlen("Empty name")+1);
			continue;
		}
		else
		{
			write(connfd, ERR_CODE, CODE_LEN);
			syslog(LOG_ERR, "Can't get phonebook entry, libgnokii error %d", error);
			return;
		}
	}
	write(connfd, END_CODE, CODE_LEN);
	
	/* Terminate phone bus */
	busterminate();
}
void get_todo()
{
	gn_todo_list todolist;
	gn_todo todo;
	gn_error error;
	int i;

	businit();
	
	/* Send confirmation to client */
	write(connfd, BEGIN_CODE, CODE_LEN);
	write(connfd, reqno, CODE_LEN);

	/* Realizing request. We are gettings entire todo list, so our 
	 * last element number is INT_MAX. When we encounter an error, we
	 * will simply break the loop */
	
	for(i = 1; i <= INT_MAX; i++)
	{
		todo.location = i;
		data->todo = &todo;
		data->todo_list = &todolist;
		error = gn_sm_functions(GN_OP_GetToDo, data, state);
		if(error == GN_ERR_NONE)
		{
			write(connfd, todo.text, sizeof(todo.text));
			usleep(100);
			if(todo.priority == GN_TODO_LOW)
				write(connfd, "low", 4);
			else if(todo.priority == GN_TODO_MEDIUM)
				write(connfd, "medium", 7);
			else if(todo.priority == GN_TODO_HIGH)
				write(connfd, "high", 5);
			else
				write(connfd, "unknown", 8);
		}
		else
			break;
	}
	write(connfd, END_CODE, CODE_LEN);
	
	busterminate();
}

void purge_todo()
{
	gn_error error;
	
	businit();

	/* Send confirmation to client */
	write(connfd, BEGIN_CODE, CODE_LEN);
	write(connfd, reqno, CODE_LEN);

	usleep(100);

	/* Realizing request */
	gn_data_clear(data);
	error = gn_sm_functions(GN_OP_DeleteAllToDos, data, state);
	if(error == GN_ERR_NONE)
		write(connfd, END_CODE, CODE_LEN);
	else
		write(connfd, ERR_CODE, CODE_LEN);
	
	busterminate();
		
}

void del_pb_entry()
{
	gn_error error;
	gn_phonebook_entry entry;
	char memtype[3];
	char pos_str[16];
	int pos;
	
	businit();

	/* Send confirmation to client */
	write(connfd, BEGIN_CODE, CODE_LEN);
	write(connfd, reqno, CODE_LEN);
	
	usleep(100);
	
	read(connfd, memtype, sizeof(memtype));
	read(connfd, pos_str, sizeof(pos_str));
	pos = atoi(pos_str);
	entry.memory_type = gn_str2memory_type(memtype);
	entry.location = pos;
	entry.empty = true;
	data->phonebook_entry = &entry;
	error = gn_sm_functions(GN_OP_DeletePhonebook, data, state);
	if(error == GN_ERR_NONE)
		write(connfd, END_CODE, CODE_LEN);
	else
		write(connfd, ERR_CODE, CODE_LEN);
	
	busterminate();
}

void write_pb_entry()
{
	char isedit[2];
	gn_error error;
	gn_phonebook_entry entry;
	char memtype[3];
	char csv[2048];
	int i;

	businit();

	/* Send confirmation to client */
	write(connfd, BEGIN_CODE, CODE_LEN);
        write(connfd, reqno, CODE_LEN);

	/* Get data from client */
	read(connfd, isedit, sizeof(isedit));
	read(connfd, memtype, sizeof(memtype));
	read(connfd, csv, sizeof(csv));
	
	memset(&entry, 0, sizeof(gn_phonebook_entry));

	entry.location = 1;
	entry.memory_type = gn_str2memory_type(memtype);

	/* Parse entry */
	error = gn_file_phonebook_raw_parse(&entry, csv);
	if(error != GN_ERR_NONE)
	{
		write(connfd, ERR_CODE, CODE_LEN);
		return;
	}
	if(strcmp(isedit, "0") == 0)
	{
		/* Find free location */
		for (i = 1; ; i++) 
		{
	           	gn_phonebook_entry aux;

			memcpy(&aux, &entry, sizeof(gn_phonebook_entry));
			data->phonebook_entry = &aux;
			data->phonebook_entry->location = i;
			error = gn_sm_functions(GN_OP_ReadPhonebook, data, state);
			if (error != GN_ERR_NONE && error != GN_ERR_EMPTYLOCATION) 
				break;
			if (aux.empty || error == GN_ERR_EMPTYLOCATION) 
			{
				entry.location = aux.location;
				error = GN_ERR_NONE;
				break;
			}
		}

	}
	
	/* Write */
	gn_phonebook_entry_sanitize(&entry);
	data->phonebook_entry = &entry;
	error = gn_sm_functions(GN_OP_WritePhonebook, data, state);
	if(error == GN_ERR_NONE)
		write(connfd, END_CODE, CODE_LEN);
	else
		write(connfd, ERR_CODE, CODE_LEN);
	
	busterminate();
}

	
main(int argc, char **argv)
{
	
	/* Checking OS name */
	if((uname(&sysname)) == -1)
		printf("Warning: Can't get name of the OS.\n");
	
	if(argc >= 2 && (strcmp(argv[1], "--nodetach") == 0))
		usleep(1);
	else
	{
		/* Daemonize the process */
		if(daemon_init() == -1)
		{
			printf("Can't daemonize process.\n");
			exit(99);
		}
	}

	/* Open syslog */
	openlog("qgnokiid", LOG_CONS, LOG_USER);
	
        /* Catch SIGTERM */
        signal(SIGTERM, got_sigterm);

	/* Create listening socket */
	if((listenfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		return(errno);
	bzero(&servaddr, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(9000);
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);

	/* Bind a name to the socket */
	if(bind(listenfd, (struct sockaddr *) &servaddr, sizeof(servaddr)) == -1)
		exit(errno);

	/* And make it listening */
	if(listen(listenfd, 5) == -1)
		exit(errno);
	for(;;)
	{
		/* Accept incoming connection */
		if((connfd = accept(listenfd, NULL, NULL)) == -1)
		{
			syslog(LOG_WARNING, "Can't accept connection");
			continue;
		}
		read(connfd, reqno, sizeof(reqno));
		if(strcmp(reqno, BEGIN_CODE) == 0)
		{
			read(connfd, reqno, sizeof(reqno));
			if(strcmp(reqno, IDENTIFY_CODE) == 0)
				phone_identify();
			else if(strcmp(reqno, SENDSMS_CODE) == 0)
				send_sms();
			else if(strcmp(reqno, GETPB_CODE) == 0)
				get_pb();
			else if(strcmp(reqno, GETTODO_CODE) == 0)
				get_todo();
			else if(strcmp(reqno, PURGETODO_CODE) == 0)
				purge_todo();
			else if(strcmp(reqno, DELPB_CODE) == 0)
				del_pb_entry();
			else if(strcmp(reqno, WRITEPB_CODE) == 0)
				write_pb_entry();
			else
				write(connfd, ERR_CODE, CODE_LEN);
		}
		else
			write(connfd, ERR_CODE, CODE_LEN);
	}
	sleep(3);
	close(connfd);
}
